@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # Urwid common display code
 #    Copyright (C) 2004-2011  Ian Ward
 #
@@ -16,24 +15,31 @@
 #    License along with this library; if not, write to the Free Software
 #    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
-# Urwid web site: http://excess.org/urwid/
+# Urwid web site: https://urwid.org/
 
-from __future__ import division, print_function
+
+from __future__ import annotations
 
 import os
 import sys
+import typing
+import warnings
 
-try:
+try:  # noqa: SIM105
     import termios
 except ImportError:
-    pass # windows
+    pass  # windows
 
-from urwid.util import StoppingContext, int_scale
 from urwid import signals
-from urwid.compat import B, bytes3, xrange, with_metaclass
+from urwid.util import StoppingContext, int_scale
+
+if typing.TYPE_CHECKING:
+    from collections.abc import Iterable, Sequence
+
+    from typing_extensions import Literal, Self
 
 # for replacing unprintable bytes with '?'
-UNPRINTABLE_TRANS_TABLE = B("?") * 32 + bytes3(list(xrange(32,256)))
+UNPRINTABLE_TRANS_TABLE = b"?" * 32 + bytes(range(32, 256))
 
 
 # signals sent by BaseScreen
@@ -42,44 +48,86 @@ INPUT_DESCRIPTORS_CHANGED = "input descriptors changed"
 
 
 # AttrSpec internal values
-_BASIC_START = 0 # first index of basic color aliases
-_CUBE_START = 16 # first index of color cube
-_CUBE_SIZE_256 = 6 # one side of the color cube
+_BASIC_START = 0  # first index of basic color aliases
+_CUBE_START = 16  # first index of color cube
+_CUBE_SIZE_256 = 6  # one side of the color cube
 _GRAY_SIZE_256 = 24
-_GRAY_START_256 = _CUBE_SIZE_256 ** 3 + _CUBE_START
-_CUBE_WHITE_256 = _GRAY_START_256 -1
+_GRAY_START_256 = _CUBE_SIZE_256**3 + _CUBE_START
+_CUBE_WHITE_256 = _GRAY_START_256 - 1
 _CUBE_SIZE_88 = 4
 _GRAY_SIZE_88 = 8
-_GRAY_START_88 = _CUBE_SIZE_88 ** 3 + _CUBE_START
-_CUBE_WHITE_88 = _GRAY_START_88 -1
+_GRAY_START_88 = _CUBE_SIZE_88**3 + _CUBE_START
+_CUBE_WHITE_88 = _GRAY_START_88 - 1
 _CUBE_BLACK = _CUBE_START
 
 # values copied from xterm 256colres.h:
-_CUBE_STEPS_256 = [0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff]
-_GRAY_STEPS_256 = [0x08, 0x12, 0x1c, 0x26, 0x30, 0x3a, 0x44, 0x4e, 0x58, 0x62,
-    0x6c, 0x76, 0x80, 0x84, 0x94, 0x9e, 0xa8, 0xb2, 0xbc, 0xc6, 0xd0,
-    0xda, 0xe4, 0xee]
+_CUBE_STEPS_256 = [0x00, 0x5F, 0x87, 0xAF, 0xD7, 0xFF]
+_GRAY_STEPS_256 = [
+    0x08,
+    0x12,
+    0x1C,
+    0x26,
+    0x30,
+    0x3A,
+    0x44,
+    0x4E,
+    0x58,
+    0x62,
+    0x6C,
+    0x76,
+    0x80,
+    0x84,
+    0x94,
+    0x9E,
+    0xA8,
+    0xB2,
+    0xBC,
+    0xC6,
+    0xD0,
+    0xDA,
+    0xE4,
+    0xEE,
+]
 # values copied from xterm 88colres.h:
-_CUBE_STEPS_88 = [0x00, 0x8b, 0xcd, 0xff]
-_GRAY_STEPS_88 = [0x2e, 0x5c, 0x73, 0x8b, 0xa2, 0xb9, 0xd0, 0xe7]
+_CUBE_STEPS_88 = [0x00, 0x8B, 0xCD, 0xFF]
+_GRAY_STEPS_88 = [0x2E, 0x5C, 0x73, 0x8B, 0xA2, 0xB9, 0xD0, 0xE7]
 # values copied from X11/rgb.txt and XTerm-col.ad:
-_BASIC_COLOR_VALUES = [(0,0,0), (205, 0, 0), (0, 205, 0), (205, 205, 0),
-    (0, 0, 238), (205, 0, 205), (0, 205, 205), (229, 229, 229),
-    (127, 127, 127), (255, 0, 0), (0, 255, 0), (255, 255, 0),
-    (0x5c, 0x5c, 0xff), (255, 0, 255), (0, 255, 255), (255, 255, 255)]
+_BASIC_COLOR_VALUES = [
+    (0, 0, 0),
+    (205, 0, 0),
+    (0, 205, 0),
+    (205, 205, 0),
+    (0, 0, 238),
+    (205, 0, 205),
+    (0, 205, 205),
+    (229, 229, 229),
+    (127, 127, 127),
+    (255, 0, 0),
+    (0, 255, 0),
+    (255, 255, 0),
+    (0x5C, 0x5C, 0xFF),
+    (255, 0, 255),
+    (0, 255, 255),
+    (255, 255, 255),
+]
 
-_COLOR_VALUES_256 = (_BASIC_COLOR_VALUES +
-    [(r, g, b) for r in _CUBE_STEPS_256 for g in _CUBE_STEPS_256
-    for b in _CUBE_STEPS_256] +
-    [(gr, gr, gr) for gr in _GRAY_STEPS_256])
-_COLOR_VALUES_88 = (_BASIC_COLOR_VALUES +
-    [(r, g, b) for r in _CUBE_STEPS_88 for g in _CUBE_STEPS_88
-    for b in _CUBE_STEPS_88] +
-    [(gr, gr, gr) for gr in _GRAY_STEPS_88])
+_COLOR_VALUES_256 = (
+    _BASIC_COLOR_VALUES
+    + [(r, g, b) for r in _CUBE_STEPS_256 for g in _CUBE_STEPS_256 for b in _CUBE_STEPS_256]
+    + [(gr, gr, gr) for gr in _GRAY_STEPS_256]
+)
+_COLOR_VALUES_88 = (
+    _BASIC_COLOR_VALUES
+    + [(r, g, b) for r in _CUBE_STEPS_88 for g in _CUBE_STEPS_88 for b in _CUBE_STEPS_88]
+    + [(gr, gr, gr) for gr in _GRAY_STEPS_88]
+)
 
-assert len(_COLOR_VALUES_256) == 256
-assert len(_COLOR_VALUES_88) == 88
+if len(_COLOR_VALUES_256) != 256:
+    raise RuntimeError(_COLOR_VALUES_256)
+if len(_COLOR_VALUES_88) != 88:
+    raise RuntimeError(_COLOR_VALUES_88)
 
+# fmt: off
 _FG_COLOR_MASK =      0x000000ffffff
 _BG_COLOR_MASK =      0xffffff000000
 _FG_BASIC_COLOR =    0x1000000000000
@@ -97,27 +145,38 @@ _BOLD =            0x400000000000000
 _BLINK =           0x800000000000000
 _ITALICS =        0x1000000000000000
 _STRIKETHROUGH =  0x2000000000000000
-_FG_MASK = (_FG_COLOR_MASK | _FG_BASIC_COLOR | _FG_HIGH_COLOR |
-    _STANDOUT | _UNDERLINE | _BLINK | _BOLD | _ITALICS | _STRIKETHROUGH)
+# fmt: on
+
+_FG_MASK = (
+    _FG_COLOR_MASK
+    | _FG_BASIC_COLOR
+    | _FG_HIGH_COLOR
+    | _STANDOUT
+    | _UNDERLINE
+    | _BLINK
+    | _BOLD
+    | _ITALICS
+    | _STRIKETHROUGH
+)
 _BG_MASK = _BG_COLOR_MASK | _BG_BASIC_COLOR | _BG_HIGH_COLOR
 
-DEFAULT = 'default'
-BLACK = 'black'
-DARK_RED = 'dark red'
-DARK_GREEN = 'dark green'
-BROWN = 'brown'
-DARK_BLUE = 'dark blue'
-DARK_MAGENTA = 'dark magenta'
-DARK_CYAN = 'dark cyan'
-LIGHT_GRAY = 'light gray'
-DARK_GRAY = 'dark gray'
-LIGHT_RED = 'light red'
-LIGHT_GREEN = 'light green'
-YELLOW = 'yellow'
-LIGHT_BLUE = 'light blue'
-LIGHT_MAGENTA = 'light magenta'
-LIGHT_CYAN = 'light cyan'
-WHITE = 'white'
+DEFAULT = "default"
+BLACK = "black"
+DARK_RED = "dark red"
+DARK_GREEN = "dark green"
+BROWN = "brown"
+DARK_BLUE = "dark blue"
+DARK_MAGENTA = "dark magenta"
+DARK_CYAN = "dark cyan"
+LIGHT_GRAY = "light gray"
+DARK_GRAY = "dark gray"
+LIGHT_RED = "light red"
+LIGHT_GREEN = "light green"
+YELLOW = "yellow"
+LIGHT_BLUE = "light blue"
+LIGHT_MAGENTA = "light magenta"
+LIGHT_CYAN = "light cyan"
+WHITE = "white"
 
 _BASIC_COLORS = [
     BLACK,
@@ -139,15 +198,16 @@ _BASIC_COLORS = [
 ]
 
 _ATTRIBUTES = {
-    'bold': _BOLD,
-    'italics': _ITALICS,
-    'underline': _UNDERLINE,
-    'blink': _BLINK,
-    'standout': _STANDOUT,
-    'strikethrough': _STRIKETHROUGH,
+    "bold": _BOLD,
+    "italics": _ITALICS,
+    "underline": _UNDERLINE,
+    "blink": _BLINK,
+    "standout": _STANDOUT,
+    "strikethrough": _STRIKETHROUGH,
 }
 
-def _value_lookup_table(values, size):
+
+def _value_lookup_table(values: Sequence[int], size: int) -> list[int]:
     """
     Generate a lookup table for finding the closest item in values.
     Lookup returns (index into values)+1
@@ -159,18 +219,18 @@ def _value_lookup_table(values, size):
     [0, 0, 0, 0, 1, 1, 1, 1, 2, 2]
     """
 
-    middle_values = [0] + [(values[i] + values[i + 1] + 1) // 2
-        for i in range(len(values) - 1)] + [size]
+    middle_values = [0] + [(values[i] + values[i + 1] + 1) // 2 for i in range(len(values) - 1)] + [size]
     lookup_table = []
-    for i in range(len(middle_values)-1):
+    for i in range(len(middle_values) - 1):
         count = middle_values[i + 1] - middle_values[i]
         lookup_table.extend([i] * count)
     return lookup_table
 
+
 _CUBE_256_LOOKUP = _value_lookup_table(_CUBE_STEPS_256, 256)
-_GRAY_256_LOOKUP = _value_lookup_table([0] + _GRAY_STEPS_256 + [0xff], 256)
+_GRAY_256_LOOKUP = _value_lookup_table([0, *_GRAY_STEPS_256, 255], 256)
 _CUBE_88_LOOKUP = _value_lookup_table(_CUBE_STEPS_88, 256)
-_GRAY_88_LOOKUP = _value_lookup_table([0] + _GRAY_STEPS_88 + [0xff], 256)
+_GRAY_88_LOOKUP = _value_lookup_table([0, *_GRAY_STEPS_88, 255], 256)
 
 # convert steps to values that will be used by string versions of the colors
 # 1 hex digit for rgb and 0..100 for grayscale
@@ -180,14 +240,10 @@ _CUBE_STEPS_88_16 = [int_scale(n, 0x100, 0x10) for n in _CUBE_STEPS_88]
 _GRAY_STEPS_88_101 = [int_scale(n, 0x100, 101) for n in _GRAY_STEPS_88]
 
 # create lookup tables for 1 hex digit rgb and 0..100 for grayscale values
-_CUBE_256_LOOKUP_16 = [_CUBE_256_LOOKUP[int_scale(n, 16, 0x100)]
-    for n in range(16)]
-_GRAY_256_LOOKUP_101 = [_GRAY_256_LOOKUP[int_scale(n, 101, 0x100)]
-    for n in range(101)]
-_CUBE_88_LOOKUP_16 = [_CUBE_88_LOOKUP[int_scale(n, 16, 0x100)]
-    for n in range(16)]
-_GRAY_88_LOOKUP_101 = [_GRAY_88_LOOKUP[int_scale(n, 101, 0x100)]
-    for n in range(101)]
+_CUBE_256_LOOKUP_16 = [_CUBE_256_LOOKUP[int_scale(n, 16, 0x100)] for n in range(16)]
+_GRAY_256_LOOKUP_101 = [_GRAY_256_LOOKUP[int_scale(n, 101, 0x100)] for n in range(101)]
+_CUBE_88_LOOKUP_16 = [_CUBE_88_LOOKUP[int_scale(n, 16, 0x100)] for n in range(16)]
+_GRAY_88_LOOKUP_101 = [_GRAY_88_LOOKUP[int_scale(n, 101, 0x100)] for n in range(101)]
 
 
 # The functions _gray_num_256() and _gray_num_88() do not include the gray
@@ -198,7 +254,7 @@ _GRAY_88_LOOKUP_101 = [_GRAY_88_LOOKUP[int_scale(n, 101, 0x100)]
 # customized by an end-user.
 
 
-def _gray_num_256(gnum):
+def _gray_num_256(gnum: int) -> int:
     """Return ths color number for gray number gnum.
 
     Color cube black and white are returned for 0 and 25 respectively
@@ -215,7 +271,7 @@ def _gray_num_256(gnum):
     return _GRAY_START_256 + gnum
 
 
-def _gray_num_88(gnum):
+def _gray_num_88(gnum: int) -> int:
     """Return ths color number for gray number gnum.
 
     Color cube black and white are returned for 0 and 9 respectively
@@ -232,11 +288,11 @@ def _gray_num_88(gnum):
     return _GRAY_START_88 + gnum
 
 
-def _color_desc_true(num):
+def _color_desc_true(num: int) -> str:
+    return f"#{num:06x}"
 
-    return "#%06x" %(num)
 
-def _color_desc_256(num):
+def _color_desc_256(num: int) -> str:
     """
     Return a string description of color number num.
     0..15 -> 'h0'..'h15' basic colors (as high-colors)
@@ -257,19 +313,20 @@ def _color_desc_256(num):
     'g11'
 
     """
-    assert num >= 0 and num < 256, num
+    if not 0 <= num < 256:
+        raise ValueError(num)
     if num < _CUBE_START:
-        return 'h%d' % num
+        return f"h{num:d}"
     if num < _GRAY_START_256:
         num -= _CUBE_START
         b, num = num % _CUBE_SIZE_256, num // _CUBE_SIZE_256
         g, num = num % _CUBE_SIZE_256, num // _CUBE_SIZE_256
         r = num % _CUBE_SIZE_256
-        return '#%x%x%x' % (_CUBE_STEPS_256_16[r], _CUBE_STEPS_256_16[g],
-            _CUBE_STEPS_256_16[b])
-    return 'g%d' % _GRAY_STEPS_256_101[num - _GRAY_START_256]
+        return f"#{_CUBE_STEPS_256_16[r]:x}{_CUBE_STEPS_256_16[g]:x}{_CUBE_STEPS_256_16[b]:x}"
+    return f"g{_GRAY_STEPS_256_101[num - _GRAY_START_256]:d}"
 
-def _color_desc_88(num):
+
+def _color_desc_88(num: int) -> str:
     """
     Return a string description of color number num.
     0..15 -> 'h0'..'h15' basic colors (as high-colors)
@@ -290,19 +347,19 @@ def _color_desc_88(num):
     'g45'
 
     """
-    assert num > 0 and num < 88
+    if not 0 < num < 88:
+        raise ValueError(num)
     if num < _CUBE_START:
-        return 'h%d' % num
+        return f"h{num:d}"
     if num < _GRAY_START_88:
         num -= _CUBE_START
         b, num = num % _CUBE_SIZE_88, num // _CUBE_SIZE_88
-        g, r= num % _CUBE_SIZE_88, num // _CUBE_SIZE_88
-        return '#%x%x%x' % (_CUBE_STEPS_88_16[r], _CUBE_STEPS_88_16[g],
-            _CUBE_STEPS_88_16[b])
-    return 'g%d' % _GRAY_STEPS_88_101[num - _GRAY_START_88]
+        g, r = num % _CUBE_SIZE_88, num // _CUBE_SIZE_88
+        return f"#{_CUBE_STEPS_88_16[r]:x}{_CUBE_STEPS_88_16[g]:x}{_CUBE_STEPS_88_16[b]:x}"
+    return f"g{_GRAY_STEPS_88_101[num - _GRAY_START_88]:d}"
 
-def _parse_color_true(desc):
 
+def _parse_color_true(desc: str) -> int | None:
     c = _parse_color_256(desc)
     if c is not None:
         (r, g, b) = _COLOR_VALUES_256[c]
@@ -313,12 +370,13 @@ def _parse_color_true(desc):
     if len(desc) == 7:
         h = desc[1:]
         return int(h, 16)
-    elif len(desc) == 4:
-        h = "0x%s0%s0%s" %(desc[1], desc[2], desc[3])
+    if len(desc) == 4:
+        h = f"0x{desc[1]}0{desc[2]}0{desc[3]}"
         return int(h, 16)
     return None
 
-def _parse_color_256(desc):
+
+def _parse_color_256(desc: str) -> int | None:
     """
     Return a color number for the description desc.
     'h0'..'h255' -> 0..255 actual color number
@@ -341,14 +399,14 @@ def _parse_color_256(desc):
         # keep the length within reason before parsing
         return None
     try:
-        if desc.startswith('h'):
+        if desc.startswith("h"):
             # high-color number
             num = int(desc[1:], 10)
             if num < 0 or num > 255:
                 return None
             return num
 
-        if desc.startswith('#') and len(desc) == 4:
+        if desc.startswith("#") and len(desc) == 4:
             # color-cube coordinates
             rgb = int(desc[1:], 16)
             if rgb < 0:
@@ -362,13 +420,13 @@ def _parse_color_256(desc):
             return _CUBE_START + (r * _CUBE_SIZE_256 + g) * _CUBE_SIZE_256 + b
 
         # Only remaining possibility is gray value
-        if desc.startswith('g#'):
+        if desc.startswith("g#"):
             # hex value 00..ff
             gray = int(desc[2:], 16)
             if gray < 0 or gray > 255:
                 return None
             gray = _GRAY_256_LOOKUP[gray]
-        elif desc.startswith('g'):
+        elif desc.startswith("g"):
             # decimal value 0..100
             gray = int(desc[1:], 10)
             if gray < 0 or gray > 100:
@@ -386,7 +444,16 @@ def _parse_color_256(desc):
     except ValueError:
         return None
 
-def _parse_color_88(desc):
+
+def _true_to_256(desc: str) -> str | None:
+    if not (desc.startswith("#") and len(desc) == 7):
+        return None
+
+    c256 = _parse_color_256("#" + "".join(format(int(x, 16) // 16, "x") for x in [desc[1:3], desc[3:5], desc[5:7]]))
+    return _color_desc_256(c256)
+
+
+def _parse_color_88(desc: str) -> int | None:
     """
     Return a color number for the description desc.
     'h0'..'h87' -> 0..87 actual color number
@@ -412,14 +479,14 @@ def _parse_color_88(desc):
         # keep the length within reason before parsing
         return None
     try:
-        if desc.startswith('h'):
+        if desc.startswith("h"):
             # high-color number
             num = int(desc[1:], 10)
             if num < 0 or num > 87:
                 return None
             return num
 
-        if desc.startswith('#') and len(desc) == 4:
+        if desc.startswith("#") and len(desc) == 4:
             # color-cube coordinates
             rgb = int(desc[1:], 16)
             if rgb < 0:
@@ -433,13 +500,13 @@ def _parse_color_88(desc):
             return _CUBE_START + (r * _CUBE_SIZE_88 + g) * _CUBE_SIZE_88 + b
 
         # Only remaining possibility is gray value
-        if desc.startswith('g#'):
+        if desc.startswith("g#"):
             # hex value 00..ff
             gray = int(desc[2:], 16)
             if gray < 0 or gray > 255:
                 return None
             gray = _GRAY_88_LOOKUP[gray]
-        elif desc.startswith('g'):
+        elif desc.startswith("g"):
             # decimal value 0..100
             gray = int(desc[1:], 10)
             if gray < 0 or gray > 100:
@@ -457,11 +524,15 @@ def _parse_color_88(desc):
     except ValueError:
         return None
 
+
 class AttrSpecError(Exception):
     pass
 
-class AttrSpec(object):
-    def __init__(self, fg, bg, colors=256):
+
+class AttrSpec:
+    __slots__ = ("__value",)
+
+    def __init__(self, fg: str, bg: str, colors: Literal[1, 16, 88, 256, 16777216] = 256) -> None:
         """
         fg -- a string containing a comma-separated foreground color
               and settings
@@ -518,63 +589,144 @@ class AttrSpec(object):
         AttrSpec('#ccc', '#000', colors=88)
         """
         if colors not in (1, 16, 88, 256, 2**24):
-            raise AttrSpecError('invalid number of colors (%d).' % colors)
-        self._value = 0 | _HIGH_88_COLOR * (colors == 88) | _HIGH_TRUE_COLOR * (colors == 2**24)
-        self.foreground = fg
-        self.background = bg
+            raise AttrSpecError(f"invalid number of colors ({colors:d}).")
+        self.__value = 0 | _HIGH_88_COLOR * (colors == 88) | _HIGH_TRUE_COLOR * (colors == 2**24)
+        self.__set_foreground(fg)
+        self.__set_background(bg)
         if self.colors > colors:
-            raise AttrSpecError(('foreground/background (%s/%s) require ' +
-                'more colors than have been specified (%d).') %
-                (repr(fg), repr(bg), colors))
+            raise AttrSpecError(
+                f"foreground/background ({fg!r}/{bg!r}) require more colors than have been specified ({colors:d})."
+            )
 
-    foreground_basic = property(lambda s: s._value & _FG_BASIC_COLOR != 0)
-    foreground_high = property(lambda s: s._value & _FG_HIGH_COLOR != 0)
-    foreground_true = property(lambda s: s._value & _FG_TRUE_COLOR != 0)
-    foreground_number = property(lambda s: s._value & _FG_COLOR_MASK)
-    background_basic = property(lambda s: s._value & _BG_BASIC_COLOR != 0)
-    background_high = property(lambda s: s._value & _BG_HIGH_COLOR != 0)
-    background_true = property(lambda s: s._value & _BG_TRUE_COLOR != 0)
-    background_number = property(lambda s: (s._value & _BG_COLOR_MASK)
-        >> _BG_SHIFT)
-    italics = property(lambda s: s._value & _ITALICS != 0)
-    bold = property(lambda s: s._value & _BOLD != 0)
-    underline = property(lambda s: s._value & _UNDERLINE != 0)
-    blink = property(lambda s: s._value & _BLINK != 0)
-    standout = property(lambda s: s._value & _STANDOUT != 0)
-    strikethrough = property(lambda s: s._value & _STRIKETHROUGH != 0)
+    def copy_modified(
+        self,
+        fg: str | None = None,
+        bg: str | None = None,
+        colors: Literal[1, 16, 88, 256, 16777216] | None = None,
+    ) -> Self:
+        if fg is None:
+            foreground = self.foreground
+        else:
+            foreground = fg
 
-    def _colors(self):
+        if bg is None:
+            background = self.background
+        else:
+            background = bg
+
+        if colors is None:
+            new_colors = self.colors
+        else:
+            new_colors = colors
+
+        return self.__class__(foreground, background, new_colors)
+
+    def __hash__(self) -> int:
+        """Instance is immutable and hashable."""
+        return hash((self.__class__, self.__value))
+
+    @property
+    def _value(self) -> int:
+        """Read-only value access."""
+        return self.__value
+
+    @property
+    def foreground_basic(self) -> bool:
+        return self.__value & _FG_BASIC_COLOR != 0
+
+    @property
+    def foreground_high(self) -> bool:
+        return self.__value & _FG_HIGH_COLOR != 0
+
+    @property
+    def foreground_true(self) -> bool:
+        return self.__value & _FG_TRUE_COLOR != 0
+
+    @property
+    def foreground_number(self) -> int:
+        return self.__value & _FG_COLOR_MASK
+
+    @property
+    def background_basic(self) -> bool:
+        return self.__value & _BG_BASIC_COLOR != 0
+
+    @property
+    def background_high(self) -> bool:
+        return self.__value & _BG_HIGH_COLOR != 0
+
+    @property
+    def background_true(self) -> bool:
+        return self.__value & _BG_TRUE_COLOR != 0
+
+    @property
+    def background_number(self) -> int:
+        return (self.__value & _BG_COLOR_MASK) >> _BG_SHIFT
+
+    @property
+    def italics(self) -> bool:
+        return self.__value & _ITALICS != 0
+
+    @property
+    def bold(self) -> bool:
+        return self.__value & _BOLD != 0
+
+    @property
+    def underline(self) -> bool:
+        return self.__value & _UNDERLINE != 0
+
+    @property
+    def blink(self) -> bool:
+        return self.__value & _BLINK != 0
+
+    @property
+    def standout(self) -> bool:
+        return self.__value & _STANDOUT != 0
+
+    @property
+    def strikethrough(self) -> bool:
+        return self.__value & _STRIKETHROUGH != 0
+
+    @property
+    def colors(self) -> int:
         """
         Return the maximum colors required for this object.
 
         Returns 256, 88, 16 or 1.
         """
-        if self._value & _HIGH_88_COLOR:
+        if self.__value & _HIGH_88_COLOR:
             return 88
-        if self._value & (_BG_HIGH_COLOR | _FG_HIGH_COLOR):
+        if self.__value & (_BG_HIGH_COLOR | _FG_HIGH_COLOR):
             return 256
-        if self._value & (_BG_TRUE_COLOR | _FG_TRUE_COLOR):
+        if self.__value & (_BG_TRUE_COLOR | _FG_TRUE_COLOR):
             return 2**24
-        if self._value & (_BG_BASIC_COLOR | _BG_BASIC_COLOR):
+        if self.__value & (_BG_BASIC_COLOR | _FG_BASIC_COLOR):
             return 16
         return 1
-    colors = property(_colors)
 
-    def __repr__(self):
+    def _colors(self) -> int:
+        warnings.warn(
+            f"Method `{self.__class__.__name__}._colors` is deprecated, "
+            f"please use property `{self.__class__.__name__}.colors`",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.colors
+
+    def __repr__(self) -> str:
         """
         Return an executable python representation of the AttrSpec
         object.
         """
-        args = "%r, %r" % (self.foreground, self.background)
+        args = f"{self.foreground!r}, {self.background!r}"
         if self.colors == 88:
             # 88-color mode is the only one that is handled differently
-            args = args + ", colors=88"
-        return "%s(%s)" % (self.__class__.__name__, args)
+            args = f"{args}, colors=88"
+        return f"{self.__class__.__name__}({args})"
 
-    def _foreground_color(self):
+    def _foreground_color(self) -> str:
         """Return only the color component of the foreground."""
         if not (self.foreground_basic or self.foreground_high or self.foreground_true):
-            return 'default'
+            return "default"
         if self.foreground_basic:
             return _BASIC_COLORS[self.foreground_number]
         if self.colors == 88:
@@ -583,89 +735,105 @@ class AttrSpec(object):
             return _color_desc_true(self.foreground_number)
         return _color_desc_256(self.foreground_number)
 
-    def _foreground(self):
-        return (self._foreground_color() +
-            ',bold' * self.bold + ',italics' * self.italics +
-            ',standout' * self.standout + ',blink' * self.blink +
-            ',underline' * self.underline + ',strikethrough' * self.strikethrough)
+    @property
+    def foreground(self) -> str:
+        return (
+            self._foreground_color()
+            + ",bold" * self.bold
+            + ",italics" * self.italics
+            + ",standout" * self.standout
+            + ",blink" * self.blink
+            + ",underline" * self.underline
+            + ",strikethrough" * self.strikethrough
+        )
 
-    def _set_foreground(self, foreground):
+    def __set_foreground(self, foreground: str) -> None:
         color = None
         flags = 0
         # handle comma-separated foreground
-        for part in foreground.split(','):
-            part = part.strip()
+        for part in foreground.split(","):
+            part = part.strip()  # noqa: PLW2901
             if part in _ATTRIBUTES:
                 # parse and store "settings"/attributes in flags
                 if flags & _ATTRIBUTES[part]:
-                    raise AttrSpecError(("Setting %s specified more than" +
-                        "once in foreground (%s)") % (repr(part),
-                        repr(foreground)))
+                    raise AttrSpecError(f"Setting {part!r} specified more than once in foreground ({foreground!r})")
                 flags |= _ATTRIBUTES[part]
                 continue
             # past this point we must be specifying a color
-            if part in ('', 'default'):
+            if part in ("", "default"):
                 scolor = 0
             elif part in _BASIC_COLORS:
                 scolor = _BASIC_COLORS.index(part)
                 flags |= _FG_BASIC_COLOR
-            elif self._value & _HIGH_88_COLOR:
+            elif self.__value & _HIGH_88_COLOR:
                 scolor = _parse_color_88(part)
                 flags |= _FG_HIGH_COLOR
-            elif self._value & _HIGH_TRUE_COLOR:
+            elif self.__value & _HIGH_TRUE_COLOR:
                 scolor = _parse_color_true(part)
                 flags |= _FG_TRUE_COLOR
             else:
-                scolor = _parse_color_256(part)
+                scolor = _parse_color_256(_true_to_256(part) or part)
                 flags |= _FG_HIGH_COLOR
             # _parse_color_*() return None for unrecognised colors
             if scolor is None:
-                raise AttrSpecError(("Unrecognised color specification %s " +
-                    "in foreground (%s)") % (repr(part), repr(foreground)))
+                raise AttrSpecError(f"Unrecognised color specification {part!r} in foreground ({foreground!r})")
             if color is not None:
-                raise AttrSpecError(("More than one color given for " +
-                    "foreground (%s)") % (repr(foreground),))
+                raise AttrSpecError(f"More than one color given for foreground ({foreground!r})")
             color = scolor
         if color is None:
             color = 0
-        self._value = (self._value & ~_FG_MASK) | color | flags
+        self.__value = (self.__value & ~_FG_MASK) | color | flags
 
-    foreground = property(_foreground, _set_foreground)
+    def _foreground(self) -> str:
+        warnings.warn(
+            f"Method `{self.__class__.__name__}._foreground` is deprecated, "
+            f"please use property `{self.__class__.__name__}.foreground`",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.foreground
 
-    def _background(self):
+    @property
+    def background(self) -> str:
         """Return the background color."""
         if not (self.background_basic or self.background_high or self.background_true):
-            return 'default'
+            return "default"
         if self.background_basic:
             return _BASIC_COLORS[self.background_number]
-        if self._value & _HIGH_88_COLOR:
+        if self.__value & _HIGH_88_COLOR:
             return _color_desc_88(self.background_number)
         if self.colors == 2**24:
             return _color_desc_true(self.background_number)
         return _color_desc_256(self.background_number)
 
-    def _set_background(self, background):
+    def __set_background(self, background: str) -> None:
         flags = 0
-        if background in ('', 'default'):
+        if background in ("", "default"):
             color = 0
         elif background in _BASIC_COLORS:
             color = _BASIC_COLORS.index(background)
             flags |= _BG_BASIC_COLOR
-        elif self._value & _HIGH_88_COLOR:
+        elif self.__value & _HIGH_88_COLOR:
             color = _parse_color_88(background)
             flags |= _BG_HIGH_COLOR
-        elif self._value & _HIGH_TRUE_COLOR:
+        elif self.__value & _HIGH_TRUE_COLOR:
             color = _parse_color_true(background)
             flags |= _BG_TRUE_COLOR
         else:
-            color = _parse_color_256(background)
+            color = _parse_color_256(_true_to_256(background) or background)
             flags |= _BG_HIGH_COLOR
         if color is None:
-            raise AttrSpecError(("Unrecognised color specification " +
-                "in background (%s)") % (repr(background),))
-        self._value = (self._value & ~_BG_MASK) | (color << _BG_SHIFT) | flags
+            raise AttrSpecError(f"Unrecognised color specification in background ({background!r})")
+        self.__value = (self.__value & ~_BG_MASK) | (color << _BG_SHIFT) | flags
 
-    background = property(_background, _set_background)
+    def _background(self) -> str:
+        warnings.warn(
+            f"Method `{self.__class__.__name__}._background` is deprecated, "
+            f"please use property `{self.__class__.__name__}.background`",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.background
 
     def get_rgb_values(self):
         """
@@ -684,42 +852,49 @@ class AttrSpec(object):
         if not (self.foreground_basic or self.foreground_high or self.foreground_true):
             vals = (None, None, None)
         elif self.colors == 88:
-            assert self.foreground_number < 88, "Invalid AttrSpec _value"
+            if self.foreground_number >= 88:
+                raise ValueError(f"Invalid AttrSpec _value: {self.foreground_number!r}")
             vals = _COLOR_VALUES_88[self.foreground_number]
         elif self.colors == 2**24:
-            h = "%06x" %(self.foreground_number)
+            h = f"{self.foreground_number:06x}"
             vals = tuple([int(x, 16) for x in [h[0:2], h[2:4], h[4:6]]])
         else:
             vals = _COLOR_VALUES_256[self.foreground_number]
 
         if not (self.background_basic or self.background_high or self.background_true):
-            return vals + (None, None, None)
-        elif self.colors == 88:
-            assert self.background_number < 88, "Invalid AttrSpec _value"
+            return (*vals, None, None, None)
+        if self.colors == 88:
+            if self.background_number >= 88:
+                raise ValueError(f"Invalid AttrSpec _value: {self.background_number!r}")
             return vals + _COLOR_VALUES_88[self.background_number]
-        elif self.colors == 2**24:
-            h = "%06x" %(self.background_number)
+        if self.colors == 2**24:
+            h = f"{self.background_number:06x}"
             return vals + tuple([int(x, 16) for x in [h[0:2], h[2:4], h[4:6]]])
-        else:
-            return vals + _COLOR_VALUES_256[self.background_number]
 
-    def __eq__(self, other):
-        return isinstance(other, AttrSpec) and self._value == other._value
+        return vals + _COLOR_VALUES_256[self.background_number]
 
-    def __ne__(self, other):
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, AttrSpec) and self.__value == other.__value
+
+    def __ne__(self, other: object) -> bool:
         return not self == other
 
-    __hash__ = object.__hash__
 
-
-class RealTerminal(object):
-    def __init__(self):
-        super(RealTerminal,self).__init__()
+class RealTerminal:
+    def __init__(self) -> None:
+        super().__init__()
         self._signal_keys_set = False
         self._old_signal_keys = None
 
-    def tty_signal_keys(self, intr=None, quit=None, start=None,
-        stop=None, susp=None, fileno=None):
+    def tty_signal_keys(
+        self,
+        intr: Literal["undefined"] | int | None = None,
+        quit: Literal["undefined"] | int | None = None,  # noqa: A002
+        start: Literal["undefined"] | int | None = None,
+        stop: Literal["undefined"] | int | None = None,
+        susp: Literal["undefined"] | int | None = None,
+        fileno: int | None = None,
+    ):
         """
         Read and/or set the tty's signal character settings.
         This function returns the current settings as a tuple.
@@ -736,29 +911,41 @@ class RealTerminal(object):
         if fileno is None:
             fileno = sys.stdin.fileno()
         if not os.isatty(fileno):
-            return
+            return None
 
         tattr = termios.tcgetattr(fileno)
         sattr = tattr[6]
-        skeys = (sattr[termios.VINTR], sattr[termios.VQUIT],
-            sattr[termios.VSTART], sattr[termios.VSTOP],
-            sattr[termios.VSUSP])
+        skeys = (
+            sattr[termios.VINTR],
+            sattr[termios.VQUIT],
+            sattr[termios.VSTART],
+            sattr[termios.VSTOP],
+            sattr[termios.VSUSP],
+        )
 
-        if intr == 'undefined': intr = 0
-        if quit == 'undefined': quit = 0
-        if start == 'undefined': start = 0
-        if stop == 'undefined': stop = 0
-        if susp == 'undefined': susp = 0
+        if intr == "undefined":
+            intr = 0
+        if quit == "undefined":
+            quit = 0  # noqa: A001
+        if start == "undefined":
+            start = 0
+        if stop == "undefined":
+            stop = 0
+        if susp == "undefined":
+            susp = 0
 
-        if intr is not None: tattr[6][termios.VINTR] = intr
-        if quit is not None: tattr[6][termios.VQUIT] = quit
-        if start is not None: tattr[6][termios.VSTART] = start
-        if stop is not None: tattr[6][termios.VSTOP] = stop
-        if susp is not None: tattr[6][termios.VSUSP] = susp
+        if intr is not None:
+            tattr[6][termios.VINTR] = intr
+        if quit is not None:
+            tattr[6][termios.VQUIT] = quit
+        if start is not None:
+            tattr[6][termios.VSTART] = start
+        if stop is not None:
+            tattr[6][termios.VSTOP] = stop
+        if susp is not None:
+            tattr[6][termios.VSUSP] = susp
 
-        if intr is not None or quit is not None or \
-            start is not None or stop is not None or \
-            susp is not None:
+        if any(item is not None for item in (intr, quit, start, stop, susp)):
             termios.tcsetattr(fileno, termios.TCSADRAIN, tattr)
             self._signal_keys_set = True
 
@@ -768,20 +955,24 @@ class RealTerminal(object):
 class ScreenError(Exception):
     pass
 
-class BaseScreen(with_metaclass(signals.MetaSignals, object)):
+
+class BaseScreen(metaclass=signals.MetaSignals):
     """
     Base class for Screen classes (raw_display.Screen, .. etc)
     """
-    signals = [UPDATE_PALETTE_ENTRY, INPUT_DESCRIPTORS_CHANGED]
 
-    def __init__(self):
-        super(BaseScreen,self).__init__()
+    signals: typing.ClassVar[list[str]] = [UPDATE_PALETTE_ENTRY, INPUT_DESCRIPTORS_CHANGED]
+
+    def __init__(self) -> None:
+        super().__init__()
         self._palette = {}
         self._started = False
 
-    started = property(lambda self: self._started)
+    @property
+    def started(self) -> bool:
+        return self._started
 
-    def start(self, *args, **kwargs):
+    def start(self, *args, **kwargs) -> StoppingContext:
         """Set up the screen.  If the screen has already been started, does
         nothing.
 
@@ -795,14 +986,14 @@ class BaseScreen(with_metaclass(signals.MetaSignals, object)):
         :meth:`_start`.
         """
         if not self._started:
+            self._started = True
             self._start(*args, **kwargs)
-        self._started = True
         return StoppingContext(self)
 
     def _start(self):
         pass
 
-    def stop(self):
+    def stop(self) -> None:
         if self._started:
             self._stop()
         self._started = False
@@ -816,16 +1007,24 @@ class BaseScreen(with_metaclass(signals.MetaSignals, object)):
 
         Deprecated in favor of calling `start` as a context manager.
         """
+        warnings.warn(
+            "run_wrapper is deprecated in favor of calling `start` as a context manager.",
+            DeprecationWarning,
+            stacklevel=3,
+        )
         with self.start(*args, **kwargs):
             return fn()
 
-
-    def register_palette(self, palette):
+    def register_palette(
+        self,
+        palette: Iterable[
+            tuple[str, str] | tuple[str, str, str] | tuple[str, str, str, str] | tuple[str, str, str, str, str, str]
+        ],
+    ) -> None:
         """Register a set of palette entries.
 
         palette -- a list of (name, like_other_name) or
-        (name, foreground, background, mono, foreground_high,
-        background_high) tuples
+        (name, foreground, background, mono, foreground_high, background_high) tuples
 
             The (name, like_other_name) format will copy the settings
             from the palette entry like_other_name, which must appear
@@ -838,19 +1037,25 @@ class BaseScreen(with_metaclass(signals.MetaSignals, object)):
         """
 
         for item in palette:
-            if len(item) in (3,4,6):
+            if len(item) in (3, 4, 6):
                 self.register_palette_entry(*item)
                 continue
             if len(item) != 2:
-                raise ScreenError("Invalid register_palette entry: %s" %
-                    repr(item))
+                raise ScreenError(f"Invalid register_palette entry: {item!r}")
             name, like_name = item
             if like_name not in self._palette:
-                raise ScreenError("palette entry '%s' doesn't exist"%like_name)
+                raise ScreenError(f"palette entry '{like_name}' doesn't exist")
             self._palette[name] = self._palette[like_name]
 
-    def register_palette_entry(self, name, foreground, background,
-        mono=None, foreground_high=None, background_high=None):
+    def register_palette_entry(
+        self,
+        name: str,
+        foreground: str,
+        background: str,
+        mono: str | None = None,
+        foreground_high: str | None = None,
+        background_high: str | None = None,
+    ) -> None:
         """Register a single palette entry.
 
         name -- new entry/attribute name
@@ -908,7 +1113,7 @@ class BaseScreen(with_metaclass(signals.MetaSignals, object)):
         """
         basic = AttrSpec(foreground, background, 16)
 
-        if type(mono) == tuple:
+        if isinstance(mono, tuple):
             # old style of specifying mono attributes was to put them
             # in a tuple.  convert to comma-separated string
             mono = ",".join(mono)
@@ -920,31 +1125,35 @@ class BaseScreen(with_metaclass(signals.MetaSignals, object)):
             foreground_high = foreground
         if background_high is None:
             background_high = background
+
+        high_256 = AttrSpec(foreground_high, background_high, 256)
         high_true = AttrSpec(foreground_high, background_high, 2**24)
 
         # 'hX' where X > 15 are different in 88/256 color, use
         # basic colors for 88-color mode if high colors are specified
         # in this way (also avoids crash when X > 87)
-        def large_h(desc):
-            if not desc.startswith('h'):
+        def large_h(desc: str) -> bool:
+            if not desc.startswith("h"):
                 return False
-            if ',' in desc:
-                desc = desc.split(',',1)[0]
+            if "," in desc:
+                desc = desc.split(",", 1)[0]
             num = int(desc[1:], 10)
             return num > 15
+
         if large_h(foreground_high) or large_h(background_high):
             high_88 = basic
         else:
             high_88 = AttrSpec(foreground_high, background_high, 88)
 
-        signals.emit_signal(self, UPDATE_PALETTE_ENTRY,
-                            name, basic, mono, high_88, high_true)
-        self._palette[name] = (basic, mono, high_88, high_true)
+        signals.emit_signal(self, UPDATE_PALETTE_ENTRY, name, basic, mono, high_88, high_256, high_true)
+        self._palette[name] = (basic, mono, high_88, high_256, high_true)
 
 
 def _test():
     import doctest
+
     doctest.testmod()
 
-if __name__=='__main__':
+
+if __name__ == "__main__":
     _test()
